import os
import cv2
import numpy as np
from torch.utils.data import Dataset
import albumentations as A
from albumentations.pytorch import ToTensorV2
import torch
import torch.nn as nn
from torchvision.models.segmentation import deeplabv3_resnet50
from torch.utils.data import DataLoader
import torch.optim as optim
from tqdm import tqdm

class CityscapesDataset(Dataset):
    def __init__(self, root_dir, split='train', transform=None):
        self.root_dir = root_dir
        self.split = split
        self.transform = transform
        self.images_dir = os.path.join(root_dir, 'leftImg8bit_trainvaltest', 'leftImg8bit', split)
        self.labels_dir = os.path.join(root_dir, 'gtFine_trainvaltest', 'gtFine', split)
        self.images = []
        self.labels = []

        for city in os.listdir(self.images_dir):
            img_dir = os.path.join(self.images_dir, city)
            lbl_dir = os.path.join(self.labels_dir, city)
            for file_name in os.listdir(img_dir):
                if file_name.endswith('_leftImg8bit.png'):
                    img_path = os.path.join(img_dir, file_name)
                    lbl_path = os.path.join(lbl_dir, file_name.replace('_leftImg8bit.png', '_gtFine_labelIds.png'))
                    self.images.append(img_path)
                    self.labels.append(lbl_path)

    def __len__(self):
        return len(self.images)

    def __getitem__(self, idx):
        img_path = self.images[idx]
        lbl_path = self.labels[idx]
        image = cv2.imread(img_path)
        label = cv2.imread(lbl_path, cv2.IMREAD_GRAYSCALE)
        if self.transform:
            augmented = self.transform(image=image, mask=label)
            image = augmented['image']
            label = augmented['mask']
        image = torch.tensor(image, dtype=torch.float32)
        label = torch.tensor(label, dtype=torch.long)
        return image, label

def compute_iou(pred, target, num_classes):
    ious = []
    pred = torch.argmax(pred, dim=1)
    for cls in range(num_classes):
        pred_inds = (pred == cls)
        target_inds = (target == cls)
        intersection = (pred_inds & target_inds).sum().item()
        union = (pred_inds | target_inds).sum().item()
        if union == 0:
            ious.append(float('nan'))  # If there is no ground truth, do not include in IoU calculation
        else:
            ious.append(intersection / union)
    return np.nanmean(ious)  # Return the mean IoU, ignoring NaN

def compute_precision_recall(pred, target, num_classes):
    pred = torch.argmax(pred, dim=1)
    precision, recall = [], []
    for cls in range(num_classes):
        pred_inds = (pred == cls)
        target_inds = (target == cls)
        true_positive = (pred_inds & target_inds).sum().item()
        predicted_positive = pred_inds.sum().item()
        actual_positive = target_inds.sum().item()
        
        if predicted_positive == 0:
            precision.append(float('nan'))
        else:
            precision.append(true_positive / predicted_positive)
        
        if actual_positive == 0:
            recall.append(float('nan'))
        else:
            recall.append(true_positive / actual_positive)
    
    return np.nanmean(precision), np.nanmean(recall)

def main():
    # Transformations
    print("Setting up transformations...")
    train_transform = A.Compose([
        A.Resize(512, 1024),
        A.HorizontalFlip(p=0.5),
        A.RandomCrop(512, 512),
        A.Normalize(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)),
        ToTensorV2()
    ])

    val_transform = A.Compose([
        A.Resize(512, 1024),
        A.Normalize(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)),
        ToTensorV2()
    ])

    # Datasets and DataLoaders
    print("Loading datasets...")
    train_dataset = CityscapesDataset(root_dir='cityscapes_subset', split='train', transform=train_transform)
    val_dataset = CityscapesDataset(root_dir='cityscapes_subset', split='val', transform=val_transform)

    train_loader = DataLoader(train_dataset, batch_size=2, shuffle=True, num_workers=4)
    val_loader = DataLoader(val_dataset, batch_size=2, shuffle=False, num_workers=4)

    # Device setup
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    print(f"Using device: {device}")

    # Model setup
    print("Setting up model...")
    model = deeplabv3_resnet50(pretrained=True)
    model.classifier[4] = nn.Conv2d(256, 19, kernel_size=(1, 1), stride=(1, 1))  # 19 classes for Cityscapes
    model = model.to(device)

    # Loss and optimizer
    criterion = nn.CrossEntropyLoss()
    optimizer = optim.Adam(model.parameters(), lr=0.001)

    num_epochs = 25
    num_classes = 19

    for epoch in range(num_epochs):
        model.train()
        running_loss = 0.0
        train_iou = 0.0
        train_precision = 0.0
        train_recall = 0.0

        print(f"Epoch {epoch+1}/{num_epochs}")

        # Training loop
        for images, labels in tqdm(train_loader, desc="Training"):
            images = images.to(device)
            labels = labels.to(device)
            optimizer.zero_grad()
            outputs = model(images)['out']
            loss = criterion(outputs, labels)
            loss.backward()
            optimizer.step()
            running_loss += loss.item()
            
            iou = compute_iou(outputs, labels, num_classes)
            precision, recall = compute_precision_recall(outputs, labels, num_classes)
            
            train_iou += iou
            train_precision += precision
            train_recall += recall
        
        train_iou /= len(train_loader)
        train_precision /= len(train_loader)
        train_recall /= len(train_loader)
        
        print(f"Training Loss: {running_loss/len(train_loader):.4f}, IoU: {train_iou:.4f}, Precision: {train_precision:.4f}, Recall: {train_recall:.4f}")

        # Validation loop
        model.eval()
        val_loss = 0.0
        val_iou = 0.0
        val_precision = 0.0
        val_recall = 0.0

        with torch.no_grad():
            for images, labels in tqdm(val_loader, desc="Validation"):
                images = images.to(device)
                labels = labels.to(device)
                outputs = model(images)['out']
                loss = criterion(outputs, labels)
                val_loss += loss.item()
                
                iou = compute_iou(outputs, labels, num_classes)
                precision, recall = compute_precision_recall(outputs, labels, num_classes)
                
                val_iou += iou
                val_precision += precision
                val_recall += recall
        
        val_iou /= len(val_loader)
        val_precision /= len(val_loader)
        val_recall /= len(val_loader)
        
        print(f"Validation Loss: {val_loss/len(val_loader):.4f}, IoU: {val_iou:.4f}, Precision: {val_precision:.4f}, Recall: {val_recall:.4f}")

if __name__ == "__main__":
    main()
